#!/usr/bin/env python
# -*- coding: utf-8 -*-
""" __init__.py
Description:
"""
__author__ = "Anthony Fong"
__copyright__ = "Copyright 2021, Anthony Fong"
__credits__ = ["Anthony Fong"]
__license__ = ""
__version__ = "1.4.0"
__maintainer__ = "Anthony Fong"
__email__ = ""
__status__ = "Prototype"

# Default Libraries #

# Downloaded Libraries #

# Local Libraries #
from .automaticproperties import AutomaticProperties
from .initmeta import InitMeta
from .wrappers import StaticWrapper, DynamicWrapper
